/** A helper class for constructing codepages.
 * Override getName() and have init() set the map array so
 * that a codepage char c equals map[c-0x80], if part of
 * the extended 8bit ascii set. 
 *
 * This class is part of Virca, 2003 Vidar Holen. 
 * Released under the GNU General Public License.
 */

import java.io.*;

abstract class MapCodepage implements Codepage {
    protected static char[] map;
    
    public String serverToUnicode(String s) {
        char[] a=s.toCharArray();
        char c;
        for(int i=0; i<a.length; i++) {
            c=a[i];
            c-=0x80;
            if(c<0x80) {
                a[i]=map[c];
            }
        }
        return new String(a);
    }
    public byte[] unicodeToServer(String s) {
        try {
            return s.getBytes(getName());
        } catch(UnsupportedEncodingException e) {
            return s.getBytes();
        }
    }
    
    // Java is too clever about converting Strings to bytes.
    public byte[] serverToServer(String s) {
        char[] c=s.toCharArray();
        byte[] b=new byte[c.length];
        for(int i=0; i<c.length; i++)
            b[i]=(byte)c[i];
        return b;
    }
}

