import java.io.*;
import java.util.*;
import javax.microedition.rms.*;
import javax.microedition.lcdui.*;

class OptionsScreen extends Form implements CommandListener {
	final static String store="profiles";
    Virca virca;
	ChoiceGroup profile,action;
	Profile[] profs;
	Profile sel;
	int profsuse=0;
    
    GlobalScreen gs;
	String name, user;
    int font, color, buffer, sync;

	
	public OptionsScreen(Virca v) {
		super(Virca.version);
		virca=v;
		loadOptions();
        gs=new GlobalScreen();
        gs.set(name,user,font,color,buffer,sync);
        gs.addCommand(new Command("Ok",Command.OK,1));
        gs.addCommand(new Command("Cancel",Command.CANCEL,2));
        gs.setCommandListener(this);
        
		profile=new ChoiceGroup("Profile",ChoiceGroup.EXCLUSIVE);
		for(int i=0; i<profsuse; i++) profile.append(profs[i].getName(),null);
		append(profile);

		addCommand(new Command("Connect",Command.ITEM,1));
		addCommand(new Command("Edit",Command.ITEM,2));
		addCommand(new Command("Delete",Command.ITEM,3));
		addCommand(new Command("Create",Command.ITEM,4));
        addCommand(new Command("Globals",Command.SCREEN,5));
        addCommand(new Command("Exit",Command.EXIT,8));
		setCommandListener(this);
	}
	
	private void ensureProfs(int n) {
		if(profs.length-profsuse<n) {
			Profile[] ps=new Profile[profs.length+n+1];
			System.arraycopy(profs,0,ps,0,profsuse);
			profs=ps;
			System.gc();
		}
	}
	
	public void commandAction(Command c, Displayable d) {
        if(d==gs) {
            if(c.getLabel().equals("Ok")) { 
                name=gs.name.getString(); 
                user=gs.user.getString(); 
                font=gs.f.getSelectedIndex();
                color=gs.colors.getSelectedIndex();
                buffer=gs.buf.getValue();
                sync=gs.syncmode.getSelectedIndex();
            } else gs.set(name,user,font,color,buffer,sync);
            
            virca.display(this);
        } else if(c.getLabel().equals("Globals")) {
            virca.display(gs);
        } else if(c.getLabel().equals("Exit")) {
			saveOptions();
			virca.shutdown();
        } else if(d==this) {
            int n=profile.getSelectedIndex();

            if(c.getLabel().equals("Connect")) {
                sel=profs[n];
                saveOptions();
                gs=null;
                System.gc();
                //quick fix to avoid stopping event thread:
                new Thread() { 
                    public void run() { virca.optsCompleted(); }  
                }.start();
            } else if(c.getLabel().equals("Edit")) {
                profs[n].set();
                virca.display(profs[n]);
            } else if(c.getLabel().equals("Delete")) {
                profile.delete(n);
                profsuse--;
                for(int i=n; i<profsuse; i++) {
                    profs[i]=profs[i+1];
                }
            } if(c.getLabel().equals("Create")) {
                InputBox.get(virca,this,"New profile","Enter the name of the new profile.","Name","");
            }
        }
				
	}

	public void inputCompleted(String s) {
		if(s==null || s.length()==0) {
			virca.display(this);
			return;
		}
		ensureProfs(1);
		profs[profsuse]=Profile.getDefault(s);
		profile.append(profs[profsuse].getName(),null);
		profs[profsuse].setOptionsScreen(this);
		virca.display(profs[profsuse]);
		profsuse++;
	}

	public void loadOptions() {
		RecordStore rs=null;
		try {
            name="";
			rs=RecordStore.openRecordStore(store,false);
			RecordEnumeration re=
				rs.enumerateRecords(null,null,false);

			profs=new Profile[re.numRecords()];
//			System.err.println("Loading "+profs.length);
			for(int i=0; i<profs.length; i++) {
				byte[] b=re.nextRecord();
				switch(b[0]) {
					case 0:
						profs[profsuse]=Profile.decode(b);
						profs[profsuse].setOptionsScreen(this);
//						System.err.println(i+"- "+profs[profsuse].getName());
						profsuse++;
						break;
					case 1:
						user=new String(b,1,b.length-1);
                        decodeGlobals(b);
						break;
				}
			}
			for(int i=0; i<profsuse; i++)
				for(int j=i; j<profsuse; j++)
					if(profs[i].getName().compareTo(profs[j].getName())>0) {
						sel=profs[i];
						profs[i]=profs[j];
						profs[j]=sel;
					}
			sel=null;
			if(profsuse==0) makeDef();
		} catch(Exception e) {
			System.err.println("Error reading configuration, will reset: "+e.getMessage());
			e.printStackTrace();
			makeDef();
		}
		try { rs.closeRecordStore(); } catch(Exception e) {}
	}
    void decodeGlobals(byte[] b) throws Exception {
		ByteArrayInputStream baos=new ByteArrayInputStream(b);
		DataInputStream dis=new DataInputStream(baos);

        dis.read();
	    name=dis.readUTF();
        user=dis.readUTF();
        font=dis.read(); 
        color=dis.read();
        buffer=dis.read();
        sync=dis.read();

        dis.close();
    }
    byte[] encodeGlobals() throws IOException {
        ByteArrayOutputStream baos=new ByteArrayOutputStream();
        DataOutputStream dos=new DataOutputStream(baos);

        dos.write(1);
        dos.writeUTF(name);
        dos.writeUTF(user);
        dos.write(font);
        dos.write(color);
        dos.write(buffer);
        dos.write(sync);
        dos.close();
        
        return baos.toByteArray();
    }

	private void makeDef() {
		profs=new Profile[1];
		profs[0]=Profile.getDefault("[Default]");
        profs[0].setOptionsScreen(this);
		profsuse=1;
		user=Profile.generateUser();
        name="";
        
        font=0;
        color=0;
        buffer=8;
        sync=0;
	}

	public void saveOptions() {
		RecordStore rs=null;
		try {
			RecordStore.deleteRecordStore(store);
		} catch(Exception e) {
			e.printStackTrace();
		}
		try {
			rs=RecordStore.openRecordStore(store,true);
			byte[] b=encodeGlobals();
			rs.addRecord(b,0,b.length);
			for(int i=0; i<profsuse; i++) {
				b=profs[i].encode();
				rs.addRecord(b,0,b.length);
			}
            System.gc();
		} catch(Exception e) {}
		try { rs.closeRecordStore(); } catch(Exception e) {}
	}

	public void profCompleted() {
		virca.display(this);
	}

	public Virca getVirca() {
		return virca;
	}

	public String getHost() { return sel.getHost(); }
	public String getNick() { return sel.getNick(); }
	public String getChan() { return sel.getChan(); }
	public int getFontSize() { return font; }
	public String getPass() { return sel.getPass(); }
    public int getSync() { return sync; }
    public int getCodepage() { return sel.getCodepage(); }
    public int getPhonepage() { return sel.getPhonepage(); }
    public int getColors() { return color; }
    public int getBuffer() { return buffer; }
	public String getUser() { return user; }
    public String getName() { return name.equals("")?Virca.version+" (VidarHolen.net)":name; }
}


class Profile extends Form implements CommandListener {
	OptionsScreen os;
	TextField h,n,c,pass;
	ChoiceGroup codepage, phonepage;
	String name;
	Command ok=new Command("Ok",Command.OK,1), cancel=new Command("Cancel",Command.CANCEL,2);

	String bh,bn,bc, bpass;
	int bcodepage, bphonepage;

	public Profile(String name, String host, String nick, String chan, String passwd, int cp, int pp) {
		super("Profile \""+name+"\"");
		this.name=name;
		h=new TextField("Host",host,40,TextField.ANY);
		n=new TextField("Nick",nick,16,TextField.ANY);
		c=new TextField("Channels",chan,96,TextField.ANY);
        pass=new TextField("Password",passwd,32,TextField.PASSWORD);

        codepage=new ChoiceGroup("Codepage",ChoiceGroup.EXCLUSIVE);
        codepage.append("None",null);
        codepage.append("Direct conversion",null);
        codepage.append("CP1251 Cyrillic",null);
        codepage.append("KOI8-R Russian",null);
        codepage.append("UTF-8",null);
        codepage.setSelectedIndex(cp,true);

        phonepage=new ChoiceGroup("Phone Input",ChoiceGroup.EXCLUSIVE);
        phonepage.append("None",null);
        phonepage.append("S55-Russian",null);
        phonepage.append("C100-Russian",null);
        phonepage.setSelectedIndex(pp,true);

       
		append(h);
		append(n);
		append(c);
        append(pass);
        append(codepage);
        append(phonepage);

		addCommand(ok);
		addCommand(cancel);
		setCommandListener(this);
		set();
	}

    //Backwards compatibility is not maintained in 1.2.0
	public static Profile decode(byte[] b) {
		try {
		ByteArrayInputStream baos=new ByteArrayInputStream(b);
		DataInputStream dis=new DataInputStream(baos);
		
        dis.read();
		String m=dis.readUTF(), h=dis.readUTF(),n=dis.readUTF(),c=dis.readUTF(), passwd=dis.readUTF();
        int cp=dis.read(), pp=dis.read();
		dis.close();
		return new Profile(m,h,n,c,passwd,cp,pp);
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public static Profile getDefault(String name) {
		Random rand=new Random();
		return new Profile(name,"irc.freenode.net:6667","Virca"+(Math.abs(rand.nextInt())%10000),"#virca","",4,0);
	}

	public byte[] encode() {
		try {
		ByteArrayOutputStream baos=new ByteArrayOutputStream();
		DataOutputStream das=new DataOutputStream(baos);
		das.write(0);
		das.writeUTF(getName());
		das.writeUTF(getHost());
		das.writeUTF(getNick());
		das.writeUTF(getChan());
        das.writeUTF(getPass());
        das.write(getCodepage());
        das.write(getPhonepage());
		das.close();
		return baos.toByteArray();
		} catch(Exception e) {
			e.printStackTrace();
			return null;
		}
	}
		
	public static String generateUser() {
		//return "v"+Integer.toHexString(Math.abs(new Random().nextInt()));
        //quakenet hack:
        Random r=new Random();
        StringBuffer s=new StringBuffer();
        s.append('v');
        for(int i=0; i<5; i++) 
            s.append((char)('a'+(Math.abs(r.nextInt())%26)));
        s.append('a');
        System.err.println(s);
        return s.toString();
	}

	public void setOptionsScreen(OptionsScreen o) { os=o; }

	public void commandAction(Command cmd, Displayable d) {
		if(cmd==cancel) {
			reset();
			os.profCompleted();
			return;
		}
		
		if(h.getString().indexOf(":")==-1)
			os.getVirca().displayError("Bad host","Hostnames must be specified in the form \"hostname:port\"",this);
		else if(n.getString().length()==0) os.getVirca().displayError("Bad nick","Please type in a nickname.",this);
		else os.profCompleted();
	}

	public String getName() { return name; }
	public String getHost() { return h.getString(); }
	public String getNick() { return n.getString(); }
	public String getChan() { return c.getString(); }
    public String getPass() { return pass.getString(); }
    public int getCodepage() { return codepage.getSelectedIndex(); }
    public int getPhonepage() { return phonepage.getSelectedIndex(); }

	public void set() {
		bh=getHost();
		bn=getNick();
		bc=getChan();
        bpass=getPass();
        bcodepage=getCodepage();
        bphonepage=getPhonepage();
	}
	public void reset() {
		h.setString(bh);
		n.setString(bn);
		c.setString(bc);
        pass.setString(bpass);
        codepage.setSelectedIndex(bcodepage,true);
        phonepage.setSelectedIndex(bphonepage,true);
	}
}

