/*
 * Virca and all its supporting classes are released under the GNU General
 * Public License, as published by the Free Software Foundation.  See
 * http://www.gnu.org/licenses/gpl.html or write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA.
 *
 *
 *  2003, 2004 Vidar Holen
 * www.vidarholen.net
 *
 */


import java.io.*;
import java.util.*;
import javax.microedition.io.*;
import javax.microedition.lcdui.*;
import javax.microedition.midlet.*;

public class Virca extends MIDlet implements Runnable, CommandListener {
    final public static String version="Virca 1.1.20"; 

    private Display disp;
    private VircaScreen vs;
    private OptionsScreen os;
    private InputScreen is;
    private QuickScreen qs;

    private volatile InputStream in;
    private OutputStream out;
    private StreamConnection conn;
    private StringBuffer sb;

    private String channels, user, name, nickname, passwd;
    private String server;
    private int syncmode;
    private String session;
    private String chantypes="#";
    private int buffer;

    private int RX=0, TX=0;
    boolean nochannels=true, sentjoins=false;
    private Codepage cp;
    private Phonepage pp;
    boolean cppp=false;
    
    protected Command input=new Command("Input",Command.OK,1), favorites=new Command("Favorites",Command.SCREEN,2), close=new Command("Close window",Command.SCREEN,3), clear=new Command("Clear window",Command.SCREEN,4), quit=new Command("Quit",Command.SCREEN,5);
    final private static int INIT=0,OPTS=1,INPUT=2,NORM=3, MENU=4, RECONN=5, SWAP=6, QUICK=7;
    private int state;

    public Virca() {
        try {
            state=INIT;
            disp=Display.getDisplay(this);

            vs=new VircaScreen();
            vs.setVirca(this);
            vs.addCommand(input);
            vs.addCommand(favorites);
            vs.addCommand(close);
            vs.addCommand(clear);
            vs.addCommand(quit);
            vs.setCommandListener(this);

            QuickScreen.load();
            qs=new QuickScreen(this);

            is=new InputScreen(this);
            os=new OptionsScreen(this);
            sb=new StringBuffer();

            setOptions();

        } catch(Exception e) {
            disp.setCurrent(new Alert(e.toString()),vs);
        }

    }

    public void commandAction(Command c, Displayable d) {
        int t;
        if(c==input) setInput();
        else if(c==favorites) setQuick();
        else if(c==close) {
            if(isChannel(vs.getCurrent())) {
                if(cppp) pSend("PART ",vs.getCurrent(),true,null,Virca.version,null);
                else send("PART "+vs.getCurrent());
            }
            vs.closeCurrent();
            setNormal();
        } else if(c==clear) {
            vs.clear();
            setNormal();
        } else if(c==quit) {
            quit();
            notifyDestroyed();
        }
    }
    protected void quit() {
        send("QUIT :"+version);
        waitAndQuit();
    }
    protected void waitAndQuit() {
        Thread t=new Thread() {
            public void run() { 
                recv();
                synchronized(this) {
                    notify();
                }
            }
        };
        synchronized(t) {
            t.start();
            try {
                t.wait(3000);
                conn.close();
            } catch(Exception e) {}
        }
        notifyDestroyed();
    }

    public void run() {
        String s=null;
        String nick=null, msg=null, action=null, target=null;
        int sys;
        int t;
        boolean me;
        try {
            while((s=recv())!=null) {
                if(s.length()==0) continue;
                try {
                    System.gc();
                    //Do pings and errors and be done with it!
                    if(s.startsWith("PING")) {
                        t=s.indexOf(":");
                        if(t!=-1) send("PONG :"+s.substring(t+1));
                        else send("PONG :"+s.substring(s.length()>5?5:4));
                        continue;
                    } else if(s.startsWith("ERROR")) {
                        vs.append(null,"! "+s,2);    
                    }

                    //Example of irc statement:
                    //:sender!user@address PRIVMSG target :message
                    //:feyd-rautha!baron@arrakis.com PRIVMSG muaddib :You smile, eh?

                    //hereby commented as
                    //:n!u@h a t :m

                    t=s.indexOf(" ");
                    nick=s.substring(1,t); //n!u@h
                    s=s.substring(t+1); //a t :m
                    t=nick.indexOf("!");
                    sys=t;
                    t=s.indexOf(" ");
                    if(t==-1) continue;
                    action=s.substring(0,t); //a
                    s=s.substring(t+1); //t :m
                    t=s.indexOf(":");
                    msg=s.substring(t+1); //m
                    if(sys==-1) {
                        if(Character.isDigit(action.charAt(0))) {
                            int num=Integer.parseInt(action);
                            switch(num) {
                            case 001: //Welcome
                            case 376: //End of MOTD (just in case no 001)
                                if(!sentjoins) {
                                    sentjoins=true;
                                    onConnect();
                                    joinChans();
                                }
                                break;
                            case 005:
                                t=s.indexOf("CHANTYPES");
                                if(t!=-1) {
                                    s=s.substring(t+10);
                                    t=s.indexOf(" ");
                                    chantypes=s.substring(0,t);
                                }
                                break;
                            case 322: //here's /list for you!!
                                t=s.indexOf(" ");
                                vs.append(null," "+s.substring(t+1),0);
                                break;
                            case 332:
                                s=s.substring(s.indexOf(" ")+1);
                                s=s.substring(0,s.indexOf(" "));
                                vs.append(s,"Topic is \""+msg+"\"",1);
                                break;
                            case 404: //moderated or not on channel
                                vs.append(null,"Can't send to channel: "+msg,0);
                                break;
                            case 401:
                                vs.append(null,"No such nick: "+msg,0);
                                break;
                            case 433:
                                vs.append(null,"This nick is taken. Change and try again.",0);
                                break;
                            case 471: //channel full
                            case 473: //invite only
                            case 474: //banned
                            case 475: //bad key
                                vs.append(null,"Unable to join: "+msg,0);
                                break;
                            case 432:
                                vs.append(null,"Bad nickname: "+msg,0);
                                break;
                            case 324:
                                msg=s.substring(s.indexOf(" ")+1);
                                s=msg.substring(0,msg.indexOf(" "));
                                msg=msg.substring(msg.indexOf(" ")+1);
                                vs.append(s,"Mode "+s+": "+msg,0);
                                break;
                            case 317: //idle time
                                String idle;
                                s=s.substring(s.indexOf(" ")+1);
                                s=s.substring(s.indexOf(" ")+1);
                                s=s.substring(0,s.indexOf(" "));
                                try {
                                    t=Integer.parseInt(s);
                                    idle=(t%60)+"s";
                                    t/=60;
                                    idle=(t%60)+"m "+idle;
                                    t/=60;
                                    idle=(t%24)+"h "+idle;
                                    t/=24;
                                    idle=t+"d "+idle;
                                    vs.append(null,"Whois info: idle "+idle,0);
                                } catch(Exception e) {
                                    vs.append(null,"Whois info: idle "+s,0);
                                }
                                break;
                            case 311: //whois user
                            case 313: //is operator?
                            case 319: //channels
                                s=s.substring(s.indexOf(" ")+1);
                                vs.append(null,"Whois info: "+s,0);
                                break;

                            case 353:
                                //damn servers:
                                //:sfrh.idlenet.org 353 koala_man @ #virca :koala_man @panda_man
                                //:leguin.freenode.net 353 panda_man = #virca :panda_man @koala_man

                                s=s.substring(s.indexOf(" ")+1); //skip nick
                                s=s.substring(s.indexOf(" ")+1); //skip [@=]
                                s=s.substring(0,s.indexOf(" "));
                                vs.append(s,"People here: "+msg,1);
                                break;
                            }
                        }

                    } else {
                        nick=nick.substring(0,sys); //n
                        t=s.indexOf(" ");
                        if(t!=-1) target=s.substring(0,t); //t
                        s=s.substring(t+1); //if a kick, this'll contain kickee

                        me=false;
                        if(action.equals("PRIVMSG") || action.equals("NOTICE")) {
                            if(msg.charAt(0)=='\001') {
                                if(action.equals("PRIVMSG")) {
                                    if(msg.startsWith("\001ACTION ")) {
                                        int end=msg.charAt(msg.length()-1)==1?
                                            msg.length()-1:msg.length();
                                        msg=msg.substring(8,end);
                                        me=true;
                                    } else if(msg.startsWith("\001VERSION")) {
                                        if(cppp) 
                                            pSend("NOTICE ",nick,true,
                                                    "\001VERSION "+version,null,"\001");
                                        else send("NOTICE "+nick+" :\001VERSION "+version+"\001");
                                        continue;
                                    } else if(msg.startsWith("\001PING ")) {
                                        if(cppp)
                                            pSend("NOTICE ",nick,true,msg,null,null);
                                        else send("NOTICE "+nick+" :"+msg);
                                        continue;
                                    }
                                } else { 
                                    if(msg.startsWith("\001PING ")) {
                                        msg=msg.substring(msg.indexOf(" ")+1);
                                        if(msg.endsWith("\001"))
                                            msg=msg.substring(0,msg.length()-1);
                                        try { 
                                            long l=System.currentTimeMillis()-Long.parseLong(msg);
                                            l/=1000;
                                            vs.append(null,"Ping reply from "+nick+": "+l+"s",0);
                                        } catch(NumberFormatException nfe) {
                                            vs.append(null,"Bad ping reply from "+nick,0);
                                        }
                                        continue;
                                    } else if(msg.startsWith("\001VERSION ")) {
                                        msg=msg.substring(msg.indexOf(" ")+1);
                                        if(msg.endsWith("\001"))
                                            msg=msg.substring(0,msg.length()-1);
                                        vs.append(null,nick+" uses "+msg,0);
                                        continue;
                                    }
                                }
                            }  
                            //if query to us. requires current nick.
                            if(ircEquals(target,nickname)) 
                                target=nick;
                            if(me) vs.append(target,"*"+nick+" "+msg,2);
                            else vs.append(target,nick+": "+msg,2);
                        } else if(action.equals("JOIN")) {
                            if(nochannels) {
                                vs.renameCurrent(msg);
                                nochannels=false;
                            }
//                            System.err.println("Received JOIN "+msg);
                            vs.append(msg,nick+" joined "+msg,1);
                        } else if(action.equals("QUIT")) {
                            //we don't keep a list, so just print to current. fixme?
                            if(isChannel(vs.getCurrent()))
                                vs.append(null, nick+" quit: "+msg,1);
                        } else if(action.equals("PART")) {
                            if(!ircEquals(nick,nickname) || vs.isOpen(target))
                                vs.append(target, nick+" left: "+msg,1);
                        } else if(action.equals("NICK")) {
                            if(ircEquals(nick,nickname)) {
                                nickname=msg;
                                vs.append(null,"Your nick is now "+msg,0);
                            } else {
                                if(!isChannel(vs.getCurrent()))
                                    vs.append(null,nick+" changed name to "+msg,0);
                            }
                        } else if(action.equals("MODE")) {
                            vs.append(target, nick+" sets mode "+msg,1);
                        } else if(action.equals("KICK")) {
                            t=s.indexOf(" ");
                            if(t==-1) continue;
                            s=s.substring(0,t);
                            vs.append(target,nick+" kicked "+s+": "+msg,1);
                        } else if(action.equals("TOPIC")) {
                            vs.append(target,nick+" set topic to: \""+msg+"\"",1);
                        }
                    }
                } catch(RuntimeException e) {
                    vs.append(null,"Hmm: "+e,0);
                    e.printStackTrace();
                }
            }
            conn.close();
        } catch(Exception e) {
            vs.append(null,"!Error: "+e+" : \""+s+"\" : "+s.length(),0);
            //			displayError("Thread error",e.toString(),vs);
            try { conn.close(); } catch(Exception e2) {}
        }
        if(state!=RECONN)
            vs.append(null,"!Disconnected.",0);
    }

    public void setOptions() {
        state=OPTS;
        disp.setCurrent(os);
    }
    public void setInput() {
        state=INPUT;
        is.init();
        is.setReply(vs.getCurrent());
        disp.setCurrent(is);
    }
    public void setNormal() {
        state=NORM;
        disp.setCurrent(vs);
    }
    public void setQuick() {
        state=QUICK;
        qs.reset();
        qs.setReply(vs.getCurrent());
        disp.setCurrent(qs);
    }

    public void optsCompleted() {
        if(state!=OPTS) return; //uh..crap
        setNormal();
        vs.init(os.getFontSize(),os.getColors(),os.getBuffer());

        vs.append(null,version,0);
        vs.append(null," Vidar Holen",0);
        vs.append(null,"www.VidarHolen.net",0);
        vs.append(null,"",0);
        vs.repaint();

        server=os.getHost();
        channels=os.getChan();
        nickname=os.getNick();
        user=os.getUser();
        name=os.getName();
        passwd=os.getPass();
        syncmode=os.getSync();
        switch(os.getCodepage()) {
            case 0: cp=null; break;  //no cp
            case 1: cp=new DirectCodepage(); break; //direct
            case 2: cp=new CP1251(); break; //cyrillic
            case 3: cp=new KOI8R(); break; //russian
            case 4: cp=new UTF8(); break; //the one true
        }
        switch(os.getPhonepage()) {
            case 0: pp=null; break; //no pp
            case 1: pp=new S55_1251(); break; //S55 Russian input
            case 2: pp=new C100_1251(); break; //C100 Russian input
        }
        if(pp!=null && !pp.isCompatible()) {
            vs.append(null,"Your phone doesn't work with the selected Phone Input.",0);
            pp=null;
        }
        
        if(cp!=null)
            cp.init();
        if(pp!=null)
            pp.init();
        
        //we'll be doing a lot of these:
        cppp=true; //(cp!=null || pp!=null);
        
        os=null;
        System.gc();

        connect();
        System.gc();
    }
    public void quickCompleted(boolean accept) {
        setNormal();
        if(accept) {
            ChoiceGroup cg=qs.getChoiceGroup();
            boolean[] b=new boolean[cg.size()];
            cg.getSelectedFlags(b);
            for(int i=0; i<b.length; i++) {
                if(b[i]) {
                    String s=cg.getString(i);
                    if(s.charAt(0)==';') s=s.substring(1);
                    inputCompleted(s,1,qs.getReply());
                }
            }
        }
    }

    public void inputCompleted(final int accept) {
        setNormal();
        if(accept>0)
            new Thread() {
            public void run() {
                inputCompleted(is.getText(),accept, is.getReply());
            }
        }.start();
    }
    public void inputCompleted(String string, int m, String rep) {
        /* In the following code you'll see a lot of if(cppp), if using
         * codepage or phonepage. The reason why the strings aren't simply
         * converted here is because they need to be sent in an odd fashion due
         * to javas "smart" string to byte conversion, leaving '?' instead of
         * unknown chars. Using a specific encoding in .getBytes unfortunatly
         * doesn't work since few phones seem to have that. If anyone could
         * FIXME, or have hints, it'd be appreciated.  
         *
         * For 1.2.0, I hope a DirectCodepage will fix things.
         */
        
        String n=string;
        if(n.length()==0) return; //nothing to see here
        int mode=m;
        int t;
        if(n.charAt(0)=='/') {
            t=n.indexOf(" ");
            String a=n.substring(1,t==-1?n.length():t).toLowerCase();
            n=t==-1?"":n.substring(t+1);
            if(a.equals("m") || a.equals("msg") || 
                    a.equals("q") || a.equals("query")) {
                t=n.indexOf(" ");
                if(t==-1) {
                    String s=phoneToUnicode(n);
                    vs.append(s,"",1);
                    vs.setScreen(s);
                } else {
                    a=n.substring(0,t);
                    n=n.substring(t+1);
                    if(cppp) {
                        pSend("PRIVMSG ",a,true,null,n,null);
                        vs.append(a,">"+nickname+": "+phoneToUnicode(n),0);
                    } else {
                        send("PRIVMSG "+a+" :"+n);
                        vs.append(a,">"+nickname+": "+n,0);
                    }
                }
            } else if(a.equals("raw")) {
                send(n);
                vs.append(null,"-->"+n,0);
            } else if(a.equals("n") || a.equals("nick")) {
                if(cppp) pSend("NICK ",null,false,null,n,null);
                else send("NICK "+n);
            } else if(a.equals("j") || a.equals("join")) {
                if(cppp) pSend("JOIN ",null,false,null,n,null);
                else send("JOIN "+n);
            } else if(a.equals("me")) {
                if(cppp) {
                    pSend("PRIVMSG ",rep,true,"\001ACTION ",n,"\001"); 
                    vs.append(rep,"* "+nickname+" "+phoneToUnicode(n),0);
                } else {
                    send("PRIVMSG "+rep+" :\001ACTION "+n+"\001");
                    vs.append(rep,"* "+nickname+" "+n,0);
                }
            } else if(a.equals("p") || a.equals("c") || a.equals("part") || a.equals("close")) {
                if(n.length()==0) n=version;
                if(cppp) pSend("PART ",rep,true,null,n,null);
                else send("PART "+rep+" :"+n);
                vs.closeCurrent();
            } else if(a.equals("quit")) {
                if(cppp) pSend("QUIT ",null,true,null,n,null);
                else send("QUIT :"+n);
                waitAndQuit();
            } else if(a.equals("w") || a.equals("where")) {
                vs.append(rep,"This is "+rep,1);
            } else if(a.equals("names")) {
                if(t==-1) {
                    if(cppp) pSend("NAMES ",rep,false,null,null,null);
                    else send("NAMES "+rep);
                } else {
                    if(cppp) pSend("NAMES ",null,false,null,n,null);
                    else send("NAMES "+n);
                }
            } else if(a.equals("whois")) {
                if(t==-1) vs.append(null,"Usage: /whois nick",0);
                else { 
                    if(cppp) pSend("WHOIS ",n,false,null,null,null);
                    else send("WHOIS "+n);
                }
            } else if(a.equals("ping")) { 
                if(cppp) pSend("PRIVMSG ",n,true,
                        "\001PING "+System.currentTimeMillis()+"\001",null,null);
                else send("PRIVMSG "+n+" :\001PING "+System.currentTimeMillis()+"\001");
            } else if(a.equals("ctcp")) { 
                t=n.indexOf(" ");
                if(cppp) 
                    pSend("PRIVMSG ",n.substring(0,t),true,"\001",n.substring(t+1),"\001");
                else send("PRIVMSG "+n.substring(0,t)+" :\001"+n.substring(t+1)+"\001");
            } else if(a.equals("t") || a.equals("topic")) {
                if(n.length()==0) vs.append(null, "Usage: /topic topic-text",0);
                else {
                    if(cppp) pSend("TOPIC ",rep,true,null,n,null);
                    else send("TOPIC "+rep+" :"+n);
                }
            } else if(a.equals("kick")) {
                t=n.indexOf(" ");
                a=n.substring(0,t);
                n=n.substring(t+1);
                //didn't see this one coming:
                if(cppp) pSend("KICK "+rep+" ",a,true,null,n,null);
                else send("KICK "+rep+" "+a+" :"+n);
            } else if(a.equals("mode")) {
                n=n.trim();
                if(n.charAt(0)!='-' && n.charAt(0)!='+')
                    send("MODE "+n);
                else send("MODE "+rep+" "+n);
            } else if(a.equals("server")) { //kinda dodgey
                if(t==-1) vs.append(null,"Usage: /server host:port",0);
                else {
                    state=RECONN;
                    server=n;
                    if(server.indexOf(":")==-1) server+=":6667";
                    send("QUIT :"+version);
                    connect();
                    state=NORM;
                }
            } else if(a.equals("data")) {
                vs.append(null,"Received "+RX+"b, Sent "+TX+"b",0);
            } else if(a.equals("date") || a.equals("time")) {
                long time=System.currentTimeMillis(), off;
                Calendar now=Calendar.getInstance();
                now.setTime(new Date(time));
                off=now.getTimeZone().getOffset(now.get(now.YEAR)>0?1:0,now.get(now.YEAR),now.get(now.MONTH),now.get(now.DATE),now.get(now.DAY_OF_WEEK),now.get(now.MILLISECOND));
//                System.err.println("Offset: "+off+", "+now.getTimeZone().getID());
                time+=off*1000;
                now.setTime(new Date(time));
                StringBuffer d=new StringBuffer("Date (");
                d.append(now.getTimeZone().getID()).append(") is ");
                d.append(now.get(now.DATE)).append("/");
                d.append(now.get(now.MONTH)+1).append(" ");
                d.append(now.get(now.YEAR)).append(", ");
                t=now.get(now.HOUR_OF_DAY);
                if(t<10) d.append(0);
                d.append(t).append(":");
                t=now.get(now.MINUTE);
                if(t<10) d.append(0);
                d.append(t).append(":");
                t=now.get(now.SECOND);
                if(t<10) d.append(0);
                d.append(t).append(".");
                vs.append(null,d.toString(),0);
            } else if(a.equals("joinall")) {
                joinChans();
            } else if(a.equals("cs")) {
                if(cppp) pSend("CHANSERV ",null,false,null,n,null);
                else send("CHANSERV "+n);
            } else if(a.equals("ns")) {
                if(cppp) pSend("NICKSERV ",null,false,null,n,null);
                else send("NICKSERV "+n);
            } else if(a.equals("ms")) {
                if(cppp) pSend("MEMOSERV ",null,false,null,n,null);
                else send("MEMOSERV "+n);
            } else if(a.equals("help")) {
                vs.append(null,"Please see www.vidarholen.net for help",0);
            } else vs.append(null,"Sorry, don't know \""+a+"\"",0);
        } else {
            if(cppp) {
                String unid=phoneToUnicode(n);
                if(mode==1) {
                    pSend("PRIVMSG ",rep,true,null,n,null);
                    vs.append(rep,">"+nickname+": "+unid,0);
                } else if(mode==2) {
                    pSend("PRIVMSG ",rep,true,"\001ACTION ",n,"\001");
                    vs.append(rep,"* "+nickname+" "+unid,0);
                }
            } else {
                if(mode==1) {
                    send("PRIVMSG "+rep+" :"+n);
                    vs.append(rep,">"+nickname+": "+n,0);
                } else if(mode==2) {
                    send("PRIVMSG "+rep+" :\001ACTION "+n+"\001");
                    vs.append(rep,"* "+nickname+" "+n,0);
                }
            }
        }
    }

    public void shutdown() {
        notifyDestroyed();
    }

    public void connect() {
        try {
            if(conn!=null) conn.close();
            if(in!=null) in.close();
            if(out!=null) out.close();
        } catch(Exception e) {}
        try {
            vs.append(null,"!Connecting... wait",0);
            conn=(StreamConnection)Connector.open("socket://"+server);
            in=conn.openInputStream();
            out=conn.openOutputStream();
            if(passwd.length()!=0) send("PASS "+passwd);

            send("USER "+user+" 0 * :"+name);
            send("NICK "+nickname);

            new Thread(this).start();
        } catch(Exception e) {
            vs.append(null,"!Error connecting: "+e.toString(),0);
            displayError("Can't connect",e.toString(),vs);
        }
    }
    private boolean isChannel(String s) {
        return chantypes.indexOf(s.substring(0,1))!=-1;
    } 
    private void joinChans() {
        int t;
        while((t=channels.indexOf(","))!=-1) {
            send("JOIN "+channels.substring(0,t));
            channels=channels.substring(t+1);
        }
        send("JOIN "+channels.substring(t+1));
    }

    private void onConnect() {
        ChoiceGroup cg=qs.getChoiceGroup();
        for(int i=0; i<cg.size(); i++) {
            String s=cg.getString(i);
            if(s.charAt(0)==';') inputCompleted(s.substring(1),0,"");
        }
    }

    public synchronized void send(String s) {
        TX+=s.length()+1;
//        System.err.println("-> "+s);
        try {
            byte[] b;
            b=s.getBytes();
            out.write(b);
            out.write('\n');
            out.flush();
        } catch(Exception e) {
            e.printStackTrace();
            vs.append(null,"!Error sending: "+e,0);
        }
        System.gc();
    }

    private String phoneToUnicode(String s) {
        if(s==null) return null;
        else return pp==null?s:pp.phoneToUnicode(s);
    }
    private byte[] phoneToServer(String s) {
        if(s==null) return null;
        if(pp!=null) s=pp.phoneToUnicode(s);
        return unicodeToServer(s);
    }
    private byte[] unicodeToServer(String s) {
        if(s==null) return null;
        if(cp!=null) return cp.unicodeToServer(s);
        else return s.getBytes();
    }
    
    //uglier than I had hoped
    /** Send code/phone-paged text.
     * @param type Type, ie "PRIVMSG "
     * @param person Unicode target, will be converted
     * @param sep Add a " :" separator.
     * @param head Header text (or null), not converted
     * @param text Phone message text, will be converted
     * @param foot Footer tet (or null, not converted
     */
    public synchronized void pSend(String type, String person, boolean sep, String head, String text, String foot) {
        try {
            byte[] cptext=phoneToServer(text), cpperson=unicodeToServer(person);
            out.write(type.getBytes());
            if(cpperson!=null) out.write(cpperson);
            if(sep) out.write(" :".getBytes());
            if(head!=null) out.write(head.getBytes());
            if(cptext!=null) out.write(cptext);
            if(foot!=null) out.write(foot.getBytes());
            out.write('\n');
            out.flush();
        } catch(Exception e) {
            e.printStackTrace();
            vs.append(null,"Error sending: "+e,0);
        }
        System.gc();
    }

    public String recv() {
        int n=0;
        try {
            if(syncmode==0) {
                while((n=in.read())!='\n') {
                    if(n==-1)
                        throw new Exception("Connection EOF@"+sb.length()+": "+RX);
//                    System.err.println("(char)"+Integer.toString(n,16)+" = "+(char)n);
                    if(n!='\r') sb.append((char)n);
                }
            } else {
                do {
                    //read if no sync or there is data
                    while(in.available()>0 && (n=in.read())!='\n') {
                        if(n==-1)
                            throw new Exception("Connection EOF@"+sb.length()+": "+RX);
                        if(n!='\r') sb.append((char)n);
                    }
                    if(n=='\n') break;
                    //time to wait for next poll
                    if(syncmode==1) Thread.sleep(200);
                } while(true); //This is what, five thousand whippings?
            }
        } catch(Exception e) {
            vs.append(null,"!Error receiving: "+e,0);
            vs.append(null,"Now have: "+sb.toString(),0);
            return null;
        }

//        System.err.println("<- "+sb);
        RX+=sb.length()+1;
        String s=sb.toString();
        if(cp!=null) s=cp.serverToUnicode(s);
        sb.delete(0,sb.length());
        return s;
    }

    //FIXME: Make proper IRC comparison.
    public static boolean ircEquals(String s, String t) {
        return (s.toLowerCase()).equals(t.toLowerCase());
    }

    protected void destroyApp(boolean uncond) {
        quit();
    }
    protected void pauseApp() {}
    protected void startApp() {}

    //Modal alert box for those big nasty messages.
    public void displayError(String err, String txt, Displayable caller) {
        Alert a=new Alert(err,txt,null,null);
        a.setTimeout(Alert.FOREVER);
        disp.setCurrent(a,caller);
    }
    public void display(Displayable d) {
        disp.setCurrent(d);
    }
}

class VircaScreen extends Canvas {
    private VircaScreenContext vsc;
    private int currentScreen;

    private Virca virca;
    private Vector al;
    private int numlines;
    private String[] lines;
    private Font font;
    int fg=0x000000, bg=0xFFFFFF;
    int[] changes;
    //fsize=font size
    //screensize=number of lines per screen
    //rotation=where in the lines[] we are
    //scroll=which screenful
    private int fsize, screensize, rotation, scroll;
    //height of channel bar, a third of that height
    private int barheight, bhd3;
    //max backbuffer
    private int maxscroll=6;
    private int maxwidth;
    boolean inited=false;

    public VircaScreen() {
        al=new Vector();
        scroll=0;
    }
    public void setVirca(Virca v) {
        virca=v;
    }
    /** Get the name of the currently displayed tab.
     */
    public String getCurrent() {
        return vsc.name;
    }
    /** Rename the current tab.
     * Could be used when someone changes nick in pm.
     */
    public void renameCurrent(String s) {
        vsc.name=s;
    }
    protected void paint(Graphics g) {
        int tmp1,tmp2;
        if(!inited) return;
        g.setColor(bg);
        g.fillRect(0,0,getWidth(),getHeight());
        g.setColor(fg);
        g.setFont(font);
        tmp1=getHeight();
        tmp2=(vsc.rotation+scroll*screensize)%vsc.lines.length;
        for(int i=0; i<screensize; i++) {
            tmp1-=fsize;
            g.drawString(vsc.lines[tmp2],0,tmp1,g.TOP|g.LEFT);
            tmp2=(tmp2+1)%vsc.lines.length;
        }
        if(scroll!=0) {
            int pos=(getHeight()/maxscroll)*scroll;
            pos=getHeight()-pos;
            g.fillRect(getWidth()-3,pos-4,3,4);
        }

        tmp1=maxwidth/al.size();
        for(int i=0; i<al.size(); i++) {
            int c=((VircaScreenContext)al.elementAt(i)).change;
            tmp2=i*tmp1;
            g.setColor(fg);
            g.drawRect(tmp2,0,tmp1,barheight);

            if(i==currentScreen) {
                g.fillRect(tmp2+tmp1/3,bhd3,tmp1/3,bhd3+1);
            } else if(c!=0) {
                g.setColor(changes[c]);
                g.fillRect(tmp2,0,tmp1,barheight);
            }
        }
    }

    /** Append text to tab.
     * The tab will be created if necessary.
     * @param target the tab name
     * @param s the text
     */
    public synchronized void append(String target, String s, int type) {
        boolean repaint=true;
        VircaScreenContext t=target==null?vsc:getTarget(target);
        if(t!=vsc) {
            if(t.change>type) repaint=false;
            else t.change=type;
        }
        StringBuffer sb=new StringBuffer();
        char[] c=s.toCharArray();
        int tmp1=0, tmp2=0, tmp3=0;

        for(int i=0; i<c.length; i++) {
            sb.append(c[i]);
            tmp3+=font.charWidth(c[i]);
            if(c[i]==' ') tmp2=i;
            if(tmp3>maxwidth) {
                if(tmp2!=0) {
                    t.add(sb.toString().substring(0,tmp2-tmp1));
                    i=tmp2;
                } else {
                    t.add(sb.toString().substring(0,sb.length()-1));
                    i--;
                }
                tmp2=0;
                tmp1=i;
                sb.delete(0,sb.length());
                sb.append(" ");
                tmp3=font.charWidth(' ');
            }
        }
        t.add(sb.toString());
        if(repaint) repaint();
    }

    /** Get or create the context for the given tab name.
     * @param s tab name
     * @return the tab's context
     */
    private VircaScreenContext getTarget(String s) {
        int tmp1;
        String o=s;
        s=s.toLowerCase();
        tmp1=vscIndex(s);
        if(tmp1==-1) {
            VircaScreenContext v=new VircaScreenContext(s,numlines);
            al.addElement(v);
            append(s,"Talking with "+o,1);
            return v;
        } else return (VircaScreenContext)al.elementAt(tmp1);
    }
    /** Find the index of a given tab's context.
     * @param s the tab name
     * @return the index of the context, or -1 if not there
     * @see getTarget(String)
     */
    private int vscIndex(String s) {
        for(int i=0; i<al.size(); i++) {
            VircaScreenContext v=(VircaScreenContext)al.elementAt(i);
            if(v.name.equals(s)) return i;
        }
        return -1;
    }

    /** Checks if there is a tab by name s currently open.
     * @param s the tab name
     * @return true if open, false if not
     */
    public boolean isOpen(String s) {
        return vscIndex(s)!=-1;
    }

    /** [Re-]calculate dimensions of this VircaScreen.
     */
    public void init() {
        maxwidth=getWidth();

        barheight=getHeight()/14;

        //round to closest 3
        barheight=(barheight/3)*3;
        //some limits though
        if(barheight<3) barheight=3;
        if(barheight>24) barheight=24;
        //we'll be doing this alot:
        bhd3=barheight/3;

        screensize=(getHeight()-barheight)/fsize;
        numlines=screensize*maxscroll;
        if(vsc==null) {
            vsc=new VircaScreenContext("!Info",numlines);
            al.addElement(vsc);
        }

        clearAll();
        inited=true;
    }
    /** Intialize screen with the given fontsize, scheme and buffersize.
     * Will call init() to recalculate due to changes in font dimensions.
     * 
     * @param fontsize the font size
     * @param colors the color scheme
     * @see init()
     */
    public void init(int fontsize, int colors, int buffer) {
        maxscroll=buffer;
        switch(fontsize) {
            case 0: fontsize=Font.SIZE_SMALL; break;
            case 1: fontsize=Font.SIZE_MEDIUM; break;
            case 2: fontsize=Font.SIZE_LARGE; break;
        }
        font=Font.getFont(Font.FACE_SYSTEM,
                Font.STYLE_PLAIN, fontsize);
        fsize=font.getHeight();
        switch(colors) {
            case 0: fg=0x000000; bg=0xffffff; 
                    changes=new int[]{bg, 0x40FF40,0x0000ff,0xFF70B0}; break;
            case 1: fg=0xffffff; bg=0x000000; 
                    changes=new int[]{bg, 0x40FF40,0x0000ff,0xFF70B0}; break;
            case 2: fg=0x000000; bg=0xFFFFFF; 
                    changes=new int[]{bg, 0xB0B0B0,0x555555,0x000000};
        }
        init();
    }

    /** Close the current window.
     * The next tab to its left will be selected.
     */
    public void closeCurrent() {
        al.removeElementAt(currentScreen);
        cycle(true);
    }

    public void keyReleased(int key) {
        int keycode=getGameAction(key);
        if(keycode==UP) scroll(true);
        else if(keycode==DOWN) scroll(false);
        else if(keycode==LEFT) cycle(true);
        else if(keycode==RIGHT) cycle(false);
        else if(keycode==FIRE) virca.setInput();
    }
    //To keep some phones from triggering shortcuts, when going out of
    //other screens, these hotkeys are on key pressed rather than released.
    public void keyPressed(int key) {
        if(key==KEY_NUM1) virca.setQuick();
        else if(key==KEY_POUND) virca.commandAction(virca.close,this);
        else if(key==KEY_STAR) { clear(); repaint(); }
    }
        

    public void pointerReleased(int x, int y) {
        int tmp1;
        y-=barheight;
        if(y<0) {
            tmp1=x*al.size()/maxwidth;
            setScreen(tmp1);

        } else {
            int thres=(getHeight()-barheight)/3;
            if(y<thres) scroll(true);
            else if(y>(getHeight()-thres)) scroll(false);
        }
    }

    /** Activate a given window.
     * @param n the index of the screen
     */
    private void setScreen(int n) {
        vsc=(VircaScreenContext)al.elementAt(n);
        vsc.change=0;
        currentScreen=n;
        scroll=0;
        repaint();
    }
    /** Activate a given window.
     * If no window by the name is found, nothing is done.
     * @param s the name of the screen
     */
    public void setScreen(String s) {
        int n=vscIndex(s.toLowerCase());
        if(n!=-1) setScreen(n);
    }

    /** Scroll one step in the given direction.
     * If there the limit of scrolling is reached, nothing is done.
     * 
     * @param up direction, true for up, false for down. 
     */
    public void scroll(boolean up) {
        if(up && scroll<(maxscroll-1)) scroll++;
        if(!up && scroll>0) scroll--;
        repaint();
    }
    /** Cycle windows one step in the given direction.
     * Trying to scroll too far results will wrap.
     * 
     * @param left the direction, true for left, false for right
     */
    public void cycle(boolean left) {
        int tmp1=currentScreen+al.size()+(left?-1:1);
        tmp1=tmp1%al.size();
        setScreen(tmp1);
    }

    /** Clear the current screen
     */
    public void clear() {
        vsc.clear();
    }
    /** Clear all screens.
     */
    public void clearAll() {
        for(int i=0; i<al.size(); i++)
            ((VircaScreenContext)al.elementAt(i)).clear();
    }

}

class VircaScreenContext {
    public String name;
    public String[] lines;
    public int rotation;
    public int change;

    public VircaScreenContext(String name, int numlines) {
        lines=new String[numlines];
        this.name=name;
        clear();
    }

    /** Clear this context.
     * System.gc() will be called to take the old lines away.
     */
    public void clear() {
        for(int i=0; i<lines.length; i++)
            lines[i]="";
        rotation=0;
        System.gc();
    }

    /** Add a line to this context.
     * The line will be added without word wrap.
     * @param s the line to add
     */
    public void add(String s) {
        rotation=(rotation+lines.length-1)%lines.length;
        lines[rotation]=s;
    }

}

class InputScreen extends TextBox implements CommandListener {
    private TextField text;
    private Virca virca;
    private String reply;
    private Command ok=new Command("Say",Command.OK,1);
    private Command emote=new Command("Emote",Command.OK,2);
    private Command cancel=new Command("Cancel",Command.CANCEL,3);
    public InputScreen(Virca v) {
        super("Input","",256,TextField.ANY);

        addCommand(ok);
        addCommand(emote);
        addCommand(cancel);
        setCommandListener(this);
        virca=v;
    }
    public void commandAction(Command c, Displayable d) {
        int n=0;
        if(c==ok) n=1;
        else if(c==emote) n=2;
        virca.inputCompleted(n);
    }
    /** Get the text the user has written.
     * @return the text the user wrote
     */
    public String getText() {
        return getString();
    }
    /** Sets the name to whom you are replying.
     * @param r the name
     */
    public void setReply(String r) {
        reply=r;
        setTitle(r);
    }

    /** Get the name of the person being adressed.
     * @return the targets name
     * @see setReply(String)
     */
    public String getReply() { return reply; }

    /** Reset the components to make way for new input.
     */
    public void init() {
        setString("");
        //        mode.setSelectedIndex(0,true);
    }
}


